/*
 * libdigidocpp
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 *
 */

#include "MACX509CertStore.h"
#include "X509CertStore_p.h"

#include "../../log.h"

#include <openssl/err.h>

#include <Security/SecCertificate.h>
#include <Security/SecTrustSettings.h>

using namespace digidoc;

/**
 * Loads all certificates from KeyChain and adds these to the certificate store.
 *
 * @throws IOException exception is throws if failed to open KeyCain
 */
MACX509CertStore::MACX509CertStore() throw(IOException)
{
    loadCerts(kSecTrustSettingsDomainSystem);
    loadCerts(kSecTrustSettingsDomainAdmin);
    INFO("Loaded %d certificates into certificate store.", sk_X509_num(d->stack));
}

/**
 * Load all certificates found in KeyChain and adds these to the cert store.
 *
 * @param trustSettings settings.
 * @throws IOException exception is throws if failed open KeyChain.
 */
void MACX509CertStore::loadCerts(unsigned int trustSettings) throw(IOException)
{
    CFArrayRef certs;
    if(SecTrustSettingsCopyCertificates(trustSettings, &certs))
        THROW_IOEXCEPTION("Failed to open KeyStore.");

    CFIndex size = CFArrayGetCount(certs);
    for(CFIndex i = 0; i < size; ++i)
    {
        SecCertificateRef cert = SecCertificateRef(CFArrayGetValueAtIndex(certs, i));
        CSSM_DATA data;
        if(SecCertificateGetData(cert, &data))
        {
            WARN("Error retrieving a CA certificate from the KeyStore");
            continue;
        }

        const unsigned char *pBytes = (const unsigned char*)data.Data;
        X509 *c = d2i_X509(0, &pBytes, data.Length);
        if(!c)
        {
            WARN("Cant add cert %ld to X509_STORE, %s", ASN1_INTEGER_get(X509_get_serialNumber(c)), ERR_reason_error_string(ERR_get_error()));
            continue;
        }

        sk_X509_push(d->stack, c);
        if(!X509_STORE_add_cert(d->store, c))
            WARN("Cant add cert %ld to X509_STORE, %s", ASN1_INTEGER_get(X509_get_serialNumber(c)), ERR_reason_error_string(ERR_get_error()));
    }
    CFRelease(certs);
}
