/*
 * libdigidocpp
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 *
 */

#include "CNGSigner.h"

#include "../../log.h"
#include "../cert/X509Cert.h"
#include "../Digest.h"
#include "../../Conf.h"
#include "../../util/File.h"

#include <algorithm>
#include <sstream>
#include <Windows.h>
#include <ncrypt.h>
#include <WinCrypt.h>
#include <cryptuiapi.h>

extern "C" {

typedef BOOL (WINAPI * PFNCCERTDISPLAYPROC)(
  __in  PCCERT_CONTEXT pCertContext,
  __in  HWND hWndSelCertDlg,
  __in  void *pvCallbackData
);

typedef struct _CRYPTUI_SELECTCERTIFICATE_STRUCT {
  DWORD               dwSize;
  HWND                hwndParent;
  DWORD               dwFlags;
  LPCWSTR             szTitle;
  DWORD               dwDontUseColumn;
  LPCWSTR             szDisplayString;
  PFNCFILTERPROC      pFilterCallback;
  PFNCCERTDISPLAYPROC pDisplayCallback;
  void *              pvCallbackData;
  DWORD               cDisplayStores;
  HCERTSTORE *        rghDisplayStores;
  DWORD               cStores;
  HCERTSTORE *        rghStores;
  DWORD               cPropSheetPages;
  LPCPROPSHEETPAGEW   rgPropSheetPages;
  HCERTSTORE          hSelectedCertStore;
} CRYPTUI_SELECTCERTIFICATE_STRUCT, *PCRYPTUI_SELECTCERTIFICATE_STRUCT;

typedef const CRYPTUI_SELECTCERTIFICATE_STRUCT
  *PCCRYPTUI_SELECTCERTIFICATE_STRUCT;

PCCERT_CONTEXT WINAPI CryptUIDlgSelectCertificateW(
  __in  PCCRYPTUI_SELECTCERTIFICATE_STRUCT pcsc
);

#define CryptUIDlgSelectCertificate CryptUIDlgSelectCertificateW

}  // extern "C"

namespace digidoc
{

typedef SECURITY_STATUS (WINAPI *s_NCryptFreeObject)( NCRYPT_HANDLE hObject );
typedef SECURITY_STATUS (WINAPI *s_NCryptSetProperty)( NCRYPT_HANDLE hObject,
    LPCWSTR pszProperty, PBYTE pbInput, DWORD cbInput, DWORD dwFlags );
typedef SECURITY_STATUS (WINAPI *s_NCryptSignHash)( NCRYPT_KEY_HANDLE hKey,
    VOID *pPaddingInfo, PBYTE pbHashValue, DWORD cbHashValue, PBYTE pbSignature,
    DWORD cbSignature, DWORD *pcbResult, DWORD dwFlags );

class CNGSignerPrivate
{
public:
    static BOOL WINAPI CertFilter(PCCERT_CONTEXT cert_context,
        BOOL* is_initial_selected_cert, void* callback_data);

    HINSTANCE h;
    s_NCryptFreeObject f_NCryptFreeObject;
    s_NCryptSetProperty f_NCryptSetProperty;
    s_NCryptSignHash f_NCryptSignHash;

    X509Cert cert;
    NCRYPT_KEY_HANDLE key;
    std::wstring pin;
    bool selectFirst;
};

}

using namespace digidoc;

BOOL CNGSignerPrivate::CertFilter(PCCERT_CONTEXT cert_context,
    BOOL *, void *callback_data)
{
    int *counter = static_cast<int*>(callback_data);
    X509Cert cert( std::vector<unsigned char>(cert_context->pbCertEncoded,
        cert_context->pbCertEncoded+cert_context->cbCertEncoded));
    std::vector<digidoc::X509Cert::KeyUsage> usage = cert.getKeyUsage();
    if( find( usage.begin(), usage.end(), digidoc::X509Cert::NonRepudiation ) != usage.end() )
    {
        (*counter)++;
        return true;
    }
    return false;
}

/**
 * Initializes CNG library
 *
 * @throws SignException exception is thrown if the provided CNG driver
 *         loading failed.
 */
CNGSigner::CNGSigner(const std::string &pin, bool selectFirst) throw(SignException)
 : d(new CNGSignerPrivate)
{
    d->h = LoadLibraryW(L"ncrypt.dll");
    d->key = 0;
    setPin(pin);
    setSelectFirst(selectFirst);
    if(!d->h)
        return;
    d->f_NCryptFreeObject = s_NCryptFreeObject(GetProcAddress(d->h, "NCryptFreeObject"));
    d->f_NCryptSetProperty = s_NCryptSetProperty(GetProcAddress(d->h, "NCryptSetProperty"));
    d->f_NCryptSignHash = s_NCryptSignHash(GetProcAddress(d->h, "NCryptSignHash"));
}

/**
 * Uninitializes CNG library and releases acquired memory.
 */
CNGSigner::~CNGSigner()
{
    if(d->h)
        FreeLibrary(d->h);
    delete d;
}

/**
 * Finds all slots connected with the computer, if the slots have tokens, lists all
 * certificates found in token. If there are more that 1 certificate lets the user
 * application select (by calling the <code>selectSignCertificate</code> callback
 * function) the certificate used for signing.
 *
 * @return returns certificate used for signing.
 * @throws throws exception if failed to select the signing certificate. For example
 *         no cards found or card has no certificate.
 */
X509* CNGSigner::getCert() const throw(SignException)
{
    if(d->cert.handle())
        return d->cert.handle();

    d->f_NCryptFreeObject( d->key );
    HCERTSTORE store = CertOpenSystemStore(0, L"MY");
    if(!store)
        return 0;

    PCCERT_CONTEXT cert_context = 0;
    if(d->selectFirst)
    {
        PCCERT_CONTEXT find = 0;
        while(find = CertFindCertificateInStore(store, X509_ASN_ENCODING|PKCS_7_ASN_ENCODING, 0, CERT_FIND_ANY, NULL, find))
        {
            BYTE keyUsage = 0;
            CertGetIntendedKeyUsage(X509_ASN_ENCODING|PKCS_7_ASN_ENCODING, find->pCertInfo, &keyUsage, 1);
            if(keyUsage & CERT_NON_REPUDIATION_KEY_USAGE)
            {
                cert_context = find;
                break;
            }
        }
    }
    else
    {
        int counter = 0;
        CRYPTUI_SELECTCERTIFICATE_STRUCT pcsc = { sizeof(pcsc) };
        pcsc.pFilterCallback = CNGSignerPrivate::CertFilter;
        pcsc.pvCallbackData = &counter;
        pcsc.cDisplayStores = 1;
        pcsc.rghDisplayStores = &store;
        cert_context = CryptUIDlgSelectCertificate(&pcsc);
    }
    if(!cert_context)
        THROW_SIGNEXCEPTION("No certificates selected");

    DWORD flags = CRYPT_ACQUIRE_ONLY_NCRYPT_KEY_FLAG|CRYPT_ACQUIRE_COMPARE_KEY_FLAG;
    //if( silent ) flags |= CRYPT_ACQUIRE_SILENT_FLAG;
    DWORD spec = 0;
    BOOL ncrypt = false;
    CryptAcquireCertificatePrivateKey( cert_context, flags, 0, &d->key, &spec, &ncrypt );
    if(!d->key)
        THROW_SIGNEXCEPTION("Failed to aquire certificate private key");

    d->cert = X509Cert( std::vector<unsigned char>(cert_context->pbCertEncoded,
        cert_context->pbCertEncoded+cert_context->cbCertEncoded));
    CertFreeCertificateContext( cert_context );

    return d->cert.handle();
}

void CNGSigner::setPin(const std::string &pin)
{
    d->pin = util::File::encodeName(pin);
}

void CNGSigner::setSelectFirst(bool first)
{
    d->selectFirst = first;
}

/**
 * Signs the digest provided using the selected certificate. If the certificate needs PIN,
 * the PIN is acquired by calling the callback function <code>getPin</code>.
 *
 * @param digest digest, which is being signed.
 * @param signature memory for the signature that is created. Struct parameter <code>length</code>
 *        is set to the actual signature length.
 * @throws SignException throws exception if the signing operation failed.
 */
void CNGSigner::sign(const Digest &digest, Signature &signature) throw(SignException)
{
    DEBUG("sign(digest = {type=%s,digest=%p,length=%d}, signature={signature=%p,length=%d})",
        OBJ_nid2sn(digest.type), digest.digest, digest.length, signature.signature, signature.length);

    if(!d->f_NCryptSignHash)
        THROW_SIGNEXCEPTION("CNG library not loaded.");

    BCRYPT_PKCS1_PADDING_INFO padInfo;
    padInfo.pszAlgId = 0;
    switch(digest.type)
    {
    case NID_sha1: padInfo.pszAlgId = NCRYPT_SHA1_ALGORITHM; break;
    case NID_sha224: padInfo.pszAlgId = L"SHA224"; break;
    case NID_sha256: padInfo.pszAlgId = NCRYPT_SHA256_ALGORITHM; break;
    case NID_sha384: padInfo.pszAlgId = NCRYPT_SHA384_ALGORITHM; break;
    case NID_sha512: padInfo.pszAlgId = NCRYPT_SHA512_ALGORITHM; break;
    case NID_md5_sha1: //padInfo.pszAlgId = L"SHAMD5"; break;
    default: break;
    }

    SECURITY_STATUS err = 0;
    if(!d->pin.empty())
        err = d->f_NCryptSetProperty(d->key, NCRYPT_PIN_PROPERTY, PBYTE(d->pin.c_str()), d->pin.size(), 0);

    err = d->f_NCryptSignHash(d->key, &padInfo, PBYTE(digest.digest), digest.length,
        signature.signature, signature.length, (DWORD*)&signature.length, BCRYPT_PAD_PKCS1);

    switch(err)
    {
    case ERROR_SUCCESS: break;
    case SCARD_W_CANCELLED_BY_USER:
    {
        SignException e(__FILE__, __LINE__, "PIN acquisition canceled.");
        e.setCode(Exception::PINCanceled);
        throw e;
        break;
    }
    default:
        std::ostringstream s;
        s << "Failed to login to token: " << err;
        SignException e(__FILE__, __LINE__, s.str());
        e.setCode(Exception::PINFailed);
        throw e;
        break;
    }
}

int digidoc::CNGSigner::type() const
{
    int result = digidoc::Digest::toMethod( Conf::getInstance()->getSignatureUri() );
    if( result == NID_sha1 )
        return result;
    bool found = false;
    std::vector<std::string> pol = digidoc::X509Cert( getCert() ).getCertificatePolicies();
    for( std::vector<std::string>::iterator i = pol.begin(); i != pol.end(); ++i )
    {
        if(i->find("1.3.6.1.4.1.10015.1.2.", 22) == 0 ||
           i->find("1.3.6.1.4.1.10015.3.2.", 22) == 0)
            found = true;
    }
    if(!found)
        return X509Cert(getCert()).getPaddingSize() > 128 ? result : NID_sha224;
    return result;
}
